"""Jieba command line interface."""

from __future__ import annotations

import sys
from argparse import ArgumentParser
from pathlib import Path

import jieba_next as jieba


def build_parser() -> ArgumentParser:
    parser = ArgumentParser(
        usage="%(prog)s [options] filename",
        description="Jieba command line interface.",
        epilog="If no filename specified, use STDIN instead.",
    )

    parser.add_argument(
        "-d",
        "--delimiter",
        metavar="DELIM",
        default=" / ",
        nargs="?",
        const=" ",
        help=(
            "use DELIM instead of ' / ' for word delimiter; "
            "or a space if it is used without DELIM"
        ),
    )
    parser.add_argument(
        "-p",
        "--pos",
        metavar="DELIM",
        nargs="?",
        const="_",
        help=(
            "enable POS tagging; if DELIM is specified, use DELIM instead of '_' "
            "for POS delimiter"
        ),
    )
    parser.add_argument("-D", "--dict", help="use DICT as dictionary")
    parser.add_argument(
        "-u",
        "--user-dict",
        help=(
            "use USER_DICT together with the default dictionary or DICT (if specified)"
        ),
    )
    parser.add_argument(
        "-a",
        "--cut-all",
        action="store_true",
        dest="cutall",
        default=False,
        help="full pattern cutting (ignored with POS tagging)",
    )
    parser.add_argument(
        "-n",
        "--no-hmm",
        dest="hmm",
        action="store_false",
        default=True,
        help="don't use the Hidden Markov Model",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        help="don't print loading messages to stderr",
    )
    parser.add_argument(
        "-V", "--version", action="version", version="Jieba " + jieba.__version__
    )
    parser.add_argument("filename", nargs="?", help="input file")
    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)

    if args.quiet:
        jieba.set_log_level(60)

    if args.dict:
        jieba.set_dictionary(args.dict)

    if args.user_dict:
        jieba.load_userdict(args.user_dict)

    if args.pos:
        import jieba_next.posseg

        pos_delim = args.pos

        def cut_generator(sentence: str):
            for w, f in jieba_next.posseg.cut(sentence, HMM=args.hmm):
                yield w + pos_delim + f

    else:

        def cut_generator(sentence: str):
            yield from jieba.cut(sentence, cut_all=args.cutall, HMM=args.hmm)

    try:
        input_file = (
            Path(args.filename).open(encoding="utf-8") if args.filename else sys.stdin
        )

        for line in input_file:
            line_content = line.removesuffix("\r\n")
            words = cut_generator(line_content)
            result = str(args.delimiter).join(words)
            print(result)
        return 0
    finally:
        if args.filename and "input_file" in locals():
            input_file.close()


if __name__ == "__main__":
    raise SystemExit(main())



MIN_FLOAT = -3.14e100
MIN_INF = float("-inf")


def get_top_states(t_state_v, K=4):
    return sorted(t_state_v, key=t_state_v.__getitem__, reverse=True)[:K]


def viterbi(obs, states, start_p, trans_p, emit_p):
    V = [{}]  # tabular
    mem_path = [{}]
    all_states = trans_p.keys()
    for y in states.get(obs[0], all_states):  # init
        V[0][y] = start_p[y] + emit_p[y].get(obs[0], MIN_FLOAT)
        mem_path[0][y] = ""
    for t in range(1, len(obs)):
        V.append({})
        mem_path.append({})
        # prev_states = get_top_states(V[t-1])
        prev_states = [x for x in mem_path[t - 1].keys() if len(trans_p[x]) > 0]

        prev_states_expect_next = {y for x in prev_states for y in trans_p[x].keys()}
        obs_states = set(states.get(obs[t], all_states)) & prev_states_expect_next

        if not obs_states:
            obs_states = prev_states_expect_next or all_states

        for y in obs_states:
            prob, state = max(
                (
                    V[t - 1][y0]
                    + trans_p[y0].get(y, MIN_INF)
                    + emit_p[y].get(obs[t], MIN_FLOAT),
                    y0,
                )
                for y0 in prev_states
            )
            V[t][y] = prob
            mem_path[t][y] = state

    last = [(V[-1][y], y) for y in mem_path[-1].keys()]
    # if len(last)==0:
    #     print obs
    prob, state = max(last)

    route = [None] * len(obs)
    i = len(obs) - 1
    while i >= 0:
        route[i] = state
        state = mem_path[i][state]
        i -= 1
    return (prob, route)



from __future__ import annotations

import logging
import marshal
import os
import re
import shutil
import sys
import tempfile
import threading
import time
import warnings
from hashlib import md5
from importlib import resources as _resources
from importlib.metadata import version as _pkg_version
from importlib.resources import files as _pkg_files
from math import log
from pathlib import Path
from typing import TYPE_CHECKING, TextIO

from . import finalseg, jieba_next_rust

if TYPE_CHECKING:
    from collections.abc import Iterator, MutableMapping, Sequence

if os.name == "nt":
    from shutil import move as _replace_file
else:
    _replace_file = os.rename


__license__ = "MIT"

try:
    __version__ = _pkg_version("jieba-next")
except Exception:  # fallback when package metadata unavailable (editable install)
    __version__ = "0.0.0"


DEFAULT_DICT = None
DEFAULT_DICT_NAME = "dict.txt"
DICT_WRITING = {}
_CACHE_ENV_VAR = "JIEBA_NEXT_CACHE_DIR"

_cache_dir_override: Path | None = None


default_logger = logging.getLogger(__name__)
default_logger.addHandler(logging.NullHandler())


pool = None

re_userdict = re.compile("^(.+?)( [0-9]+)?( [a-z]+)?$", re.UNICODE)
re_eng = re.compile("[a-zA-Z0-9]", re.UNICODE)

# \u4E00-\u9FD5a-zA-Z0-9+#&\._ : All non-space characters. Will be handled with re_han
# \r\n|\s : whitespace characters. Will not be handled.
re_han_default = re.compile("([\u4e00-\u9fd5a-zA-Z0-9+#&\\._%]+)", re.UNICODE)
re_skip_default = re.compile("(\r\n|\\s)", re.UNICODE)
re_han_cut_all = re.compile("([\u4e00-\u9fd5]+)", re.UNICODE)
re_skip_cut_all = re.compile("[^a-zA-Z0-9+#\n]", re.UNICODE)


def _user_cache_dir(app_name: str, app_author: str | None = None) -> str:
    """Return a per-user cache directory path cross-platform without external deps.

    Rough logic:
    * Windows: %LOCALAPPDATA%/<AppName> or %APPDATA% fallback.
    * macOS: ~/Library/Caches/<AppName>
    * Linux/Unix: $XDG_CACHE_HOME/<AppName> or ~/.cache/<AppName>
    """
    name = app_name or "jieba-next"
    if os.name == "nt":  # Windows
        base = (
            os.environ.get("LOCALAPPDATA")
            or os.environ.get("APPDATA")
            or tempfile.gettempdir()
        )
        return str(Path(base) / name)
    if sys.platform == "darwin":  # macOS
        return str(Path.home() / "Library" / "Caches" / name)
    xdg = os.environ.get("XDG_CACHE_HOME")  # Linux / other Unix
    if xdg:
        return str(Path(xdg) / name)
    return str(Path.home() / ".cache" / name)


def configure_logging(level: int | str = logging.INFO, *, stream=None) -> None:
    """Configure a basic stream handler for jieba_next (idempotent)."""
    if not any(isinstance(h, logging.StreamHandler) for h in default_logger.handlers):
        handler = logging.StreamHandler(stream or sys.stderr)
        formatter = logging.Formatter("%(levelname)s %(name)s: %(message)s")
        handler.setFormatter(formatter)
        default_logger.addHandler(handler)
    default_logger.setLevel(level)


def enable_default_logging() -> None:
    """Enable INFO level logging with a basic handler if none configured."""
    configure_logging(logging.INFO)


def setLogLevel(log_level):
    """Deprecated. Use set_log_level.

    Retained for compatibility with jieba/jieba_fast.
    """
    warnings.warn(
        "setLogLevel is deprecated, use set_log_level instead",
        DeprecationWarning,
        stacklevel=2,
    )
    default_logger.setLevel(log_level)


def set_log_level(log_level):
    """Set logging level for jieba_next's default logger."""
    configure_logging(log_level)


class JiebaError(Exception):
    """Base exception for jieba_next."""


class DictionaryFormatError(JiebaError):
    """Raised when dictionary file has invalid formatting."""


class DictionaryNotFoundError(JiebaError):
    """Raised when specified dictionary path does not exist."""


def set_cache_dir(path: str | Path) -> None:
    """Override the cache directory used for prefix dict caches."""
    global _cache_dir_override
    p = Path(path).expanduser().resolve()
    p.mkdir(parents=True, exist_ok=True)
    _cache_dir_override = p
    default_logger.debug("Cache directory overridden: %s", p)


def get_cache_dir() -> Path:
    if _cache_dir_override is not None:
        return _cache_dir_override
    env_dir = os.environ.get(_CACHE_ENV_VAR)
    if env_dir:
        p = Path(env_dir).expanduser()
        p.mkdir(parents=True, exist_ok=True)
        return p
    if _user_cache_dir is not None:
        base = Path(_user_cache_dir("jieba-next", "jieba-next"))
        base.mkdir(parents=True, exist_ok=True)
        return base
    # fallback
    fallback = Path(tempfile.gettempdir()) / "jieba-next-cache"
    fallback.mkdir(parents=True, exist_ok=True)
    return fallback


def _cache_file_name(abs_path: str | Path | None) -> str:
    if not abs_path or abs_path == DEFAULT_DICT:
        return "jieba-next.cache"
    abs_str = str(abs_path)
    return f"jieba-next.u{md5(abs_str.encode('utf-8', 'replace')).hexdigest()}.cache"


def _resolve_cache_file(abs_path: str | Path | None) -> Path:
    return get_cache_dir() / _cache_file_name(abs_path)


def open_dict_resource(dictionary: str | Path | None) -> TextIO:
    """Open dictionary path or built-in resource and return a text stream."""
    if dictionary in (None, DEFAULT_DICT):
        dict_path = _pkg_files(__package__).joinpath(DEFAULT_DICT_NAME)
        return Path(dict_path).open(encoding="utf-8")
    p = Path(dictionary).expanduser()
    if not p.is_file():
        raise DictionaryNotFoundError(f"dictionary file does not exist: {p}")
    return p.open(encoding="utf-8")


class Tokenizer:
    def __init__(self, dictionary: str | Path | None = DEFAULT_DICT):
        self.lock = threading.RLock()
        if dictionary == DEFAULT_DICT:
            self.dictionary = dictionary
        else:
            self.dictionary = Path(dictionary).resolve()
        self.FREQ = {}
        self.total = 0
        self.user_word_tag_tab = {}
        self.initialized = False
        self.tmp_dir = None
        self.cache_file = None
        self._rust_prefix = None  # fast trie for DAG+DP

    def __repr__(self):
        return f"<Tokenizer dictionary={self.dictionary!r}>"

    def gen_pfdict(self, f: TextIO) -> tuple[dict[str, int], int]:
        lfreq: dict[str, int] = {}
        ltotal: int = 0
        f_name = getattr(f, "name", "stream")
        for lineno, line in enumerate(f, 1):
            line_parts = line.strip().split(" ")
            if len(line_parts) < 2 or not line_parts[1].isdigit():
                raise DictionaryFormatError(
                    f"invalid dictionary entry in {f_name} at Line {lineno}: {line}"
                )
            word, freq = line_parts[:2]
            freq = int(freq)
            lfreq[word] = freq
            ltotal += freq
            for ch in range(len(word)):
                wfrag = word[: ch + 1]
                if wfrag not in lfreq:
                    lfreq[wfrag] = 0
        f.close()
        return lfreq, ltotal

    def initialize(self, dictionary: str | Path | None = None) -> None:
        if dictionary:
            abs_path = Path(dictionary).resolve()
            if self.dictionary == abs_path and self.initialized:
                return
            else:
                self.dictionary = abs_path
                self.initialized = False
        else:
            abs_path = self.dictionary

        with self.lock:
            try:
                with DICT_WRITING[abs_path]:
                    pass
            except KeyError:
                pass
            if self.initialized:
                return

            default_logger.debug(
                "Building prefix dict from %s ...", abs_path or "the default dictionary"
            )
            t1 = time.time()
            cache_file = (
                Path(self.cache_file)
                if self.cache_file
                else _resolve_cache_file(abs_path)
            )
            tmpdir = cache_file.parent

            load_from_cache_fail = True
            if cache_file.is_file() and (
                abs_path == DEFAULT_DICT
                or cache_file.stat().st_mtime > Path(abs_path).stat().st_mtime
            ):
                default_logger.debug("Loading model cache: %s", cache_file)
                try:
                    with cache_file.open("rb") as cf:
                        self.FREQ, self.total = marshal.load(cf)
                    load_from_cache_fail = False
                except Exception:
                    default_logger.warning("Failed to load cache, rebuilding.")
                    load_from_cache_fail = True

            if load_from_cache_fail:
                wlock = DICT_WRITING.get(abs_path, threading.RLock())
                DICT_WRITING[abs_path] = wlock
                with wlock:
                    self.FREQ, self.total = self.gen_pfdict(self.get_dict_file())
                    default_logger.debug("Writing model cache: %s", cache_file)
                    try:
                        # prevent moving across different filesystems
                        fd, fpath = tempfile.mkstemp(dir=tmpdir)
                        with os.fdopen(fd, "wb") as temp_cache_file:
                            marshal.dump((self.FREQ, self.total), temp_cache_file)
                        _replace_file(fpath, cache_file)
                    except Exception:
                        default_logger.exception("Failed persisting cache file")

                try:
                    del DICT_WRITING[abs_path]
                except KeyError:
                    pass

            self.initialized = True
            # Build fast Rust prefix dict once dictionary is ready
            try:
                self._rust_prefix = jieba_next_rust.PrefixDict(
                    self.FREQ, float(self.total)
                )
            except Exception:
                self._rust_prefix = None
            default_logger.info(
                "Loaded prefix dict in %.3fs (entries=%d)",
                time.time() - t1,
                len(self.FREQ),
            )
            default_logger.debug("Prefix dict built successfully")

    def check_initialized(self) -> None:
        if not self.initialized:
            self.initialize()

    def calc(
        self,
        sentence: str,
        DAG: dict[int, list[int]],
        route: MutableMapping[int, tuple[float, int]],
    ) -> None:
        N: int = len(sentence)
        route[N] = (0, 0)
        logtotal = log(self.total)
        for idx in range(N - 1, -1, -1):
            route[idx] = max(
                (
                    log(self.FREQ.get(sentence[idx : x + 1]) or 1)
                    - logtotal
                    + route[x + 1][0],
                    x,
                )
                for x in DAG[idx]
            )

    def get_DAG(self, sentence: str) -> dict[int, list[int]]:
        self.check_initialized()
        DAG: dict[int, list[int]] = {}
        N: int = len(sentence)
        for k in range(N):
            tmplist = []
            i = k
            frag = sentence[k]
            while i < N and frag in self.FREQ:
                if self.FREQ[frag]:
                    tmplist.append(i)
                i += 1
                frag = sentence[k : i + 1]
            if not tmplist:
                tmplist.append(k)
            DAG[k] = tmplist
        return DAG

    def __cut_all(self, sentence: str) -> Iterator[str]:
        dag = self.get_DAG(sentence)
        old_j = -1
        for k, L in dag.items():
            if len(L) == 1 and k > old_j:
                yield sentence[k : L[0] + 1]
                old_j = L[0]
            else:
                for j in L:
                    if j > k:
                        yield sentence[k : j + 1]
                        old_j = j

    def __cut_DAG_NO_HMM(self, sentence: str) -> Iterator[str]:
        self.check_initialized()
        if self._rust_prefix is None:
            self._ensure_rust_prefix()
        if self._rust_prefix is not None:
            route = list(self._rust_prefix.get_dag_and_calc(sentence))
        else:
            route = []
            jieba_next_rust._get_DAG_and_calc(
                self.FREQ, sentence, route, float(self.total)
            )
        x = 0
        N = len(sentence)
        buf = ""
        while x < N:
            y = route[x] + 1
            l_word = sentence[x:y]
            if re_eng.match(l_word) and len(l_word) == 1:
                buf += l_word
                x = y
            else:
                if buf:
                    yield buf
                    buf = ""
                yield l_word
                x = y
        if buf:
            yield buf
            buf = ""

    def __cut_DAG(self, sentence: str) -> Iterator[str]:
        self.check_initialized()
        if self._rust_prefix is None:
            self._ensure_rust_prefix()
        if self._rust_prefix is not None:
            route = list(self._rust_prefix.get_dag_and_calc(sentence))
        else:
            route = []
            jieba_next_rust._get_DAG_and_calc(
                self.FREQ, sentence, route, float(self.total)
            )

        x = 0
        buf = ""
        N = len(sentence)
        while x < N:
            y = route[x] + 1
            l_word = sentence[x:y]
            if y - x == 1:
                buf += l_word
            else:
                if buf:
                    if len(buf) == 1:
                        yield buf
                        buf = ""
                    else:
                        if not self.FREQ.get(buf):
                            recognized = finalseg.cut(buf)
                            for t in recognized:
                                yield t
                        else:
                            for elem in buf:
                                yield elem
                        buf = ""
                yield l_word
            x = y

        if buf:
            if len(buf) == 1:
                yield buf
            elif not self.FREQ.get(buf):
                recognized = finalseg.cut(buf)
                for t in recognized:
                    yield t
            else:
                for elem in buf:
                    yield elem

    def cut(
        self, sentence: str, cut_all: bool = False, HMM: bool = True
    ) -> Iterator[str]:
        """
        The main function that segments an entire sentence that contains
        Chinese characters into seperated words.

        Parameter:
            - sentence: The str to be segmented.
            - cut_all: Model type. True for full pattern, False for accurate pattern.
            - HMM: Whether to use the Hidden Markov Model.
        """
        if cut_all:
            re_han = re_han_cut_all
            re_skip = re_skip_cut_all
        else:
            re_han = re_han_default
            re_skip = re_skip_default
        if cut_all:
            cut_block = self.__cut_all
        elif HMM:
            cut_block = self.__cut_DAG
        else:
            cut_block = self.__cut_DAG_NO_HMM
        blocks = re_han.split(sentence)
        for blk in blocks:
            if not blk:
                continue
            if re_han.match(blk):
                yield from cut_block(blk)
            else:
                tmp = re_skip.split(blk)
                for x in tmp:
                    if re_skip.match(x):
                        yield x
                    elif not cut_all:
                        yield from x
                    else:
                        yield x

    def cut_for_search(self, sentence: str, HMM: bool = True) -> Iterator[str]:
        """
        Finer segmentation for search engines.
        """
        words = self.cut(sentence, HMM=HMM)
        for w in words:
            if len(w) > 2:
                for i in range(len(w) - 1):
                    gram2 = w[i : i + 2]
                    if self.FREQ.get(gram2):
                        yield gram2
            if len(w) > 3:
                for i in range(len(w) - 2):
                    gram3 = w[i : i + 3]
                    if self.FREQ.get(gram3):
                        yield gram3
            yield w

    def lcut(self, *args, **kwargs) -> list[str]:
        return list(self.cut(*args, **kwargs))

    def lcut_for_search(self, *args, **kwargs) -> list[str]:
        return list(self.cut_for_search(*args, **kwargs))

    _lcut = lcut
    _lcut_for_search = lcut_for_search

    def _lcut_no_hmm(self, sentence: str) -> list[str]:
        return self.lcut(sentence, False, False)

    def _lcut_all(self, sentence: str) -> list[str]:
        return self.lcut(sentence, True)

    def _lcut_for_search_no_hmm(self, sentence: str) -> list[str]:
        return self.lcut_for_search(sentence, False)

    def get_dict_file(self) -> TextIO:
        return open_dict_resource(self.dictionary)

    def load_userdict(self, f: str | Path | TextIO) -> None:
        """
        Load personalized dict to improve detect rate.

        Parameter:
            - f : A plain text file contains words and their ocurrences.
                  Can be a file-like object, or the path of the dictionary file,
                  whose encoding must be utf-8.

        Structure of dict file:
        word1 freq1 word_type1
        word2 freq2 word_type2
        ...
        Word type may be ignored
        """
        self.check_initialized()
        if isinstance(f, str):
            f = Path(f).open(encoding="utf-8")

        for ln in f:
            line = ln.strip()
            if not line:
                continue
            # match won't be None because there's at least one character
            word, freq, tag = re_userdict.match(line).groups()
            if freq is not None:
                freq = freq.strip()
            if tag is not None:
                tag = tag.strip()
            self.add_word(word, freq, tag)

    def add_word(
        self, word: str, freq: int | None = None, tag: str | None = None
    ) -> None:
        """
        Add a word to dictionary.

        freq and tag can be omitted, freq defaults to be a calculated value
        that ensures the word can be cut out.
        """
        self.check_initialized()
        freq = int(freq) if freq is not None else self.suggest_freq(word, False)
        self.FREQ[word] = freq
        self.total += freq
        if tag:
            self.user_word_tag_tab[word] = tag
        for ch in range(len(word)):
            wfrag = word[: ch + 1]
            if wfrag not in self.FREQ:
                self.FREQ[wfrag] = 0
        if freq == 0:
            finalseg.add_force_split(word)
        # PrefixDict becomes stale; rebuild lazily next initialize
        self._rust_prefix = None

    def del_word(self, word: str) -> None:
        """
        Convenient function for deleting a word.
        """
        self.add_word(word, 0)
        self._rust_prefix = None

    def suggest_freq(self, segment: str | Sequence[str], tune: bool = False) -> int:
        """
        Suggest word frequency to force the characters in a word to be
        joined or splitted.

        Parameter:
            - segment : The segments that the word is expected to be cut into,
                        If the word should be treated as a whole, use a str.
            - tune : If True, tune the word frequency.

        Note that HMM may affect the final result. If the result doesn't change,
        set HMM=False.
        """
        self.check_initialized()
        ftotal = float(self.total)
        freq = 1
        if isinstance(segment, str):
            word = segment
            for seg in self.cut(word, HMM=False):
                freq *= self.FREQ.get(seg, 1) / ftotal
            freq = max(int(freq * self.total) + 1, self.FREQ.get(word, 1))
        else:
            segment = tuple(map(str, segment))
            word = "".join(segment)
            for seg in segment:
                freq *= self.FREQ.get(seg, 1) / ftotal
            freq = min(int(freq * self.total), self.FREQ.get(word, 0))
        if tune:
            add_word(word, freq)
        return freq

    def tokenize(
        self, unicode_sentence: str, mode: str = "default", HMM: bool = True
    ) -> Iterator[tuple[str, int, int]]:
        """
        Tokenize a sentence and yields tuples of (word, start, end)

        Parameter:
            - sentence: the str to be segmented.
            - mode: "default" or "search", "search" is for finer segmentation.
            - HMM: whether to use the Hidden Markov Model.
        """
        start = 0
        if mode == "default":
            for w in self.cut(unicode_sentence, HMM=HMM):
                width = len(w)
                yield (w, start, start + width)
                start += width
        else:
            for w in self.cut(unicode_sentence, HMM=HMM):
                width = len(w)
                if len(w) > 2:
                    for i in range(len(w) - 1):
                        gram2 = w[i : i + 2]
                        if self.FREQ.get(gram2):
                            yield (gram2, start + i, start + i + 2)
                if len(w) > 3:
                    for i in range(len(w) - 2):
                        gram3 = w[i : i + 3]
                        if self.FREQ.get(gram3):
                            yield (gram3, start + i, start + i + 3)
                yield (w, start, start + width)
                start += width

    def set_dictionary(self, dictionary_path: str | Path) -> None:
        with self.lock:
            abs_path = Path(dictionary_path).resolve()
            if not Path(abs_path).is_file():
                raise DictionaryNotFoundError(
                    f"jieba_next: file does not exist: {abs_path}"
                )
            self.dictionary = abs_path
            self.initialized = False
            self._rust_prefix = None

    def _ensure_rust_prefix(self) -> None:
        """Rebuild Rust PrefixDict from current FREQ/total if missing."""
        if self.initialized and self._rust_prefix is None:
            try:
                self._rust_prefix = jieba_next_rust.PrefixDict(
                    self.FREQ, float(self.total)
                )
            except Exception:
                self._rust_prefix = None


# default Tokenizer instance

dt = Tokenizer()


# global functions
def get_freq(key: str, default: int | None = None) -> int | None:
    """Get word frequency from the in-memory dictionary.

    Preferred new name. Returns `default` when key not found.
    """
    return dt.FREQ.get(key, default)


def get_FREQ(k: str, d: int | None = None) -> int | None:
    """Deprecated alias of get_freq."""
    warnings.warn(
        "get_FREQ is deprecated, use get_freq instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return dt.FREQ.get(k, d)


add_word = dt.add_word
calc = dt.calc
cut = dt.cut
cut_for_search = dt.cut_for_search
del_word = dt.del_word
get_DAG = dt.get_DAG
get_dict_file = dt.get_dict_file
initialize = dt.initialize
lcut = dt.lcut
lcut_for_search = dt.lcut_for_search
load_userdict = dt.load_userdict
set_dictionary = dt.set_dictionary
suggest_freq = dt.suggest_freq
tokenize = dt.tokenize
user_word_tag_tab = dt.user_word_tag_tab


def _replace_file(src, dest):
    # rename can't be used across different file systems
    shutil.copy(src, dest)
    Path(src).unlink()


# Explicit public API
__all__ = [
    # Core types/instances
    "Tokenizer",
    "dt",
    # Logging
    "set_log_level",
    "setLogLevel",
    "configure_logging",
    "enable_default_logging",
    # Core operations
    "add_word",
    "calc",
    "cut",
    "cut_for_search",
    "del_word",
    "get_DAG",
    "get_dict_file",
    "get_freq",
    "get_FREQ",
    "initialize",
    "lcut",
    "lcut_for_search",
    "load_userdict",
    "set_dictionary",
    "suggest_freq",
    "tokenize",
    "user_word_tag_tab",
    # Cache helpers
    "set_cache_dir",
    "get_cache_dir",
    "open_dict_resource",
    # Exceptions
    "JiebaError",
    "DictionaryFormatError",
    "DictionaryNotFoundError",
    # Metadata
    "__version__",
    "__license__",
]



from __future__ import annotations

import re
from typing import TYPE_CHECKING

import jieba_next

from .char_state_tab import P as char_state_tab_P
from .prob_emit import P as emit_P
from .prob_start import P as start_P
from .prob_trans import P as trans_P
from .viterbi import viterbi

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


re_han_detail = re.compile(r"([\u4E00-\u9FD5]+)")
re_skip_detail = re.compile(r"([\.0-9]+|[a-zA-Z0-9]+)")
re_han_internal = re.compile(r"([\u4E00-\u9FD5a-zA-Z0-9+#&\._]+)")
re_skip_internal = re.compile(r"(\r\n|\s)")

re_eng = re.compile(r"[a-zA-Z0-9]+")
re_num = re.compile(r"[\.0-9]+")

re_eng1 = re.compile(r"^[a-zA-Z0-9]$", re.UNICODE)


class Pair:
    """词与词性二元组。保持与原始 jieba 兼容的可迭代/比较行为。"""

    word: str
    flag: str

    def __init__(self, word: str, flag: str):
        self.word = word
        self.flag = flag

    def __unicode__(self) -> str:
        return f"{self.word}/{self.flag}"

    def __repr__(self) -> str:
        return f"pair({self.word!r}, {self.flag!r})"

    def __str__(self) -> str:
        return self.__unicode__()

    def __iter__(self):  # -> Iterator[str]
        return iter((self.word, self.flag))

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Pair):
            return NotImplemented
        return self.word < other.word

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Pair)
            and self.word == other.word
            and self.flag == other.flag
        )

    def __hash__(self) -> int:
        return hash(self.word)

    def encode(self, arg: str) -> bytes:
        return self.__unicode__().encode(arg)


class POSTokenizer:
    tokenizer: jieba_next.Tokenizer
    word_tag_tab: dict[str, str]

    def __init__(self, tokenizer: jieba_next.Tokenizer | None = None):
        self.tokenizer = tokenizer or jieba_next.Tokenizer()
        self.load_word_tag(self.tokenizer.get_dict_file())

    def __repr__(self) -> str:
        return f"<POSTokenizer tokenizer={self.tokenizer!r}>"

    def __getattr__(self, name: str):
        if name in ("cut_for_search", "lcut_for_search", "tokenize"):
            # may be possible?
            raise NotImplementedError
        return getattr(self.tokenizer, name)

    def initialize(self, dictionary: str | Path | None = None) -> None:
        self.tokenizer.initialize(dictionary)
        self.load_word_tag(self.tokenizer.get_dict_file())

    def load_word_tag(self, f) -> None:  # TextIO but keep loose for compatibility
        self.word_tag_tab = {}
        f_name = getattr(f, "name", "stream")
        for lineno, line in enumerate(f, 1):
            try:
                line = line.strip()
                if not line:
                    continue
                word, _, tag = line.split(" ")
                self.word_tag_tab[word] = tag
            except Exception as e:
                raise ValueError(
                    f"invalid POS dictionary entry in {f_name} at Line {lineno}: {line}"
                ) from e
        f.close()

    def makesure_userdict_loaded(self) -> None:
        if self.tokenizer.user_word_tag_tab:
            self.word_tag_tab.update(self.tokenizer.user_word_tag_tab)
            self.tokenizer.user_word_tag_tab = {}

    def __cut(self, sentence: str) -> Iterator[Pair]:
        prob, pos_list = viterbi(sentence, char_state_tab_P, start_P, trans_P, emit_P)
        begin, nexti = 0, 0

        for i, char in enumerate(sentence):
            pos = pos_list[i][0]
            if pos == "B":
                begin = i
            elif pos == "E":
                yield Pair(sentence[begin : i + 1], pos_list[i][1])
                nexti = i + 1
            elif pos == "S":
                yield Pair(char, pos_list[i][1])
                nexti = i + 1
        if nexti < len(sentence):
            yield Pair(sentence[nexti:], pos_list[nexti][1])

    def __cut_detail(self, sentence: str) -> Iterator[Pair]:
        blocks = re_han_detail.split(sentence)
        for blk in blocks:
            if re_han_detail.match(blk):
                yield from self.__cut(blk)
            else:
                tmp = re_skip_detail.split(blk)
                for x in tmp:
                    if x:
                        if re_num.match(x):
                            yield Pair(x, "m")
                        elif re_eng.match(x):
                            yield Pair(x, "eng")
                        else:
                            yield Pair(x, "x")

    def __cut_DAG_NO_HMM(self, sentence: str) -> Iterator[Pair]:
        DAG = self.tokenizer.get_DAG(sentence)
        route = {}
        self.tokenizer.calc(sentence, DAG, route)
        x = 0
        N = len(sentence)
        buf = ""
        while x < N:
            y = route[x][1] + 1
            l_word = sentence[x:y]
            if re_eng1.match(l_word):
                buf += l_word
                x = y
            else:
                if buf:
                    yield Pair(buf, "eng")
                    buf = ""
                yield Pair(l_word, self.word_tag_tab.get(l_word, "x"))
                x = y
        if buf:
            yield Pair(buf, "eng")
            buf = ""

    def __cut_DAG(self, sentence: str) -> Iterator[Pair]:
        DAG = self.tokenizer.get_DAG(sentence)
        route = {}

        self.tokenizer.calc(sentence, DAG, route)

        x = 0
        buf = ""
        N = len(sentence)
        while x < N:
            y = route[x][1] + 1
            l_word = sentence[x:y]
            if y - x == 1:
                buf += l_word
            else:
                if buf:
                    if len(buf) == 1:
                        yield Pair(buf, self.word_tag_tab.get(buf, "x"))
                    elif not self.tokenizer.FREQ.get(buf):
                        recognized = self.__cut_detail(buf)
                        for t in recognized:
                            yield t
                    else:
                        for elem in buf:
                            yield Pair(elem, self.word_tag_tab.get(elem, "x"))
                    buf = ""
                yield Pair(l_word, self.word_tag_tab.get(l_word, "x"))
            x = y

        if buf:
            if len(buf) == 1:
                yield Pair(buf, self.word_tag_tab.get(buf, "x"))
            elif not self.tokenizer.FREQ.get(buf):
                recognized = self.__cut_detail(buf)
                for t in recognized:
                    yield t
            else:
                for elem in buf:
                    yield Pair(elem, self.word_tag_tab.get(elem, "x"))

    def __cut_internal(self, sentence: str, HMM: bool = True) -> Iterator[Pair]:
        self.makesure_userdict_loaded()
        blocks = re_han_internal.split(sentence)
        if HMM:
            cut_blk = self.__cut_DAG
        else:
            cut_blk = self.__cut_DAG_NO_HMM

        for blk in blocks:
            if re_han_internal.match(blk):
                yield from cut_blk(blk)
            else:
                tmp = re_skip_internal.split(blk)
                for x in tmp:
                    if re_skip_internal.match(x):
                        yield Pair(x, "x")
                    else:
                        for xx in x:
                            word_tag = self.word_tag_tab.get(xx)
                            if word_tag:
                                yield Pair(xx, word_tag)
                            else:
                                yield Pair(xx, "x")

    def _lcut_internal(self, sentence: str) -> list[Pair]:
        return list(self.__cut_internal(sentence))

    def _lcut_internal_no_hmm(self, sentence: str) -> list[Pair]:
        return list(self.__cut_internal(sentence, False))

    def cut(self, sentence: str, HMM: bool = True):  # -> Iterator[Pair]
        self.makesure_userdict_loaded()

        if HMM:
            cuter = self.__cut
        else:
            cuter = self.tokenizer.cut

        if HMM:
            blocks = re_han_detail.split(sentence)
        else:
            blocks = jieba_next.re_han_default.split(sentence)

        for blk in blocks:
            if (HMM and re_han_detail.match(blk)) or (
                not HMM and jieba_next.re_han_default.match(blk)
            ):
                for word in cuter(blk):
                    if isinstance(word, Pair):
                        yield word
                    else:
                        for xx in word:
                            word_tag = self.word_tag_tab.get(xx)
                            if word_tag:
                                yield Pair(xx, word_tag)
                            else:
                                yield Pair(xx, "x")
            else:
                tmp = re_skip_detail.split(blk)
                for x in tmp:
                    if x:
                        if re_num.match(x):
                            yield Pair(x, "m")
                        elif re_eng.match(x):
                            yield Pair(x, "eng")
                        else:
                            yield Pair(x, "x")

    def lcut(self, *args, **kwargs) -> list[Pair]:
        return list(self.cut(*args, **kwargs))


# default Tokenizer instance


dt: POSTokenizer = POSTokenizer(jieba_next.dt)

# global functions
cut = dt.cut
lcut = dt.lcut
_lcut = dt._lcut
_lcut_no_hmm = dt._lcut_no_hmm
initialize = dt.initialize


# Explicit public API for posseg
__all__ = [
    "Pair",
    "POSTokenizer",
    "dt",
    "cut",
    "lcut",
    "initialize",
]



from .textrank import TextRank
from .tfidf import TFIDF

try:
    from .analyzer import ChineseAnalyzer
except ImportError:
    pass

default_tfidf = TFIDF()
default_textrank = TextRank()

extract_tags = tfidf = default_tfidf.extract_tags
set_idf_path = default_tfidf.set_idf_path
textrank = default_textrank.extract_tags


def set_stop_words(stop_words_path):
    default_tfidf.set_stop_words(stop_words_path)
    default_textrank.set_stop_words(stop_words_path)


__all__ = [
    "ChineseAnalyzer",
    "TFIDF",
    "TextRank",
    "extract_tags",
    "tfidf",
    "set_idf_path",
    "textrank",
    "set_stop_words",
]



from __future__ import annotations

from operator import itemgetter
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

import jieba_next
import jieba_next.posseg

DEFAULT_IDF = Path(__file__).parent / "idf.txt"


class KeywordExtractor:
    STOP_WORDS = {
        "the",
        "of",
        "is",
        "and",
        "to",
        "in",
        "that",
        "we",
        "for",
        "an",
        "are",
        "by",
        "be",
        "as",
        "on",
        "with",
        "can",
        "if",
        "from",
        "which",
        "you",
        "it",
        "this",
        "then",
        "at",
        "have",
        "all",
        "not",
        "one",
        "has",
        "or",
    }

    def set_stop_words(self, stop_words_path: str | Path) -> None:
        abs_path = Path(stop_words_path).resolve()
        if not Path(abs_path).is_file():
            raise Exception("jieba_next: file does not exist: " + abs_path)
        with abs_path.open(encoding="utf-8") as f:
            for line in f:
                self.stop_words.add(line.strip())

    def extract_tags(self, *args, **kwargs):
        raise NotImplementedError


class IDFLoader:
    path: str
    idf_freq: dict[str, float]
    median_idf: float

    def __init__(self, idf_path: str | Path | None = None):
        self.path = ""
        self.idf_freq = {}
        self.median_idf = 0.0
        if idf_path:
            self.set_new_path(idf_path)

    def set_new_path(self, new_idf_path: str | Path) -> None:
        if self.path != new_idf_path:
            self.path = new_idf_path
            with Path(new_idf_path).open(encoding="utf-8") as f:
                self.idf_freq = {}
                for line in f:
                    word, freq = line.strip().split(" ")
                    self.idf_freq[word] = float(freq)
            self.median_idf = sorted(self.idf_freq.values())[len(self.idf_freq) // 2]

    def get_idf(self) -> tuple[dict[str, float], float]:
        return self.idf_freq, self.median_idf


class TFIDF(KeywordExtractor):
    tokenizer: jieba_next.Tokenizer
    postokenizer: jieba_next.posseg.POSTokenizer
    stop_words: set[str]
    idf_loader: IDFLoader
    idf_freq: dict[str, float]
    median_idf: float

    def __init__(self, idf_path: str | Path | None = None):
        self.tokenizer = jieba_next.dt
        self.postokenizer = jieba_next.posseg.dt
        self.stop_words = self.STOP_WORDS.copy()
        self.idf_loader = IDFLoader(idf_path or DEFAULT_IDF)
        self.idf_freq, self.median_idf = self.idf_loader.get_idf()

    def set_idf_path(self, idf_path: str | Path) -> None:
        new_abs_path = Path(idf_path).resolve()
        if not Path(new_abs_path).is_file():
            raise Exception("jieba_next: file does not exist: " + new_abs_path)
        self.idf_loader.set_new_path(new_abs_path)
        self.idf_freq, self.median_idf = self.idf_loader.get_idf()

    def extract_tags(
        self,
        sentence: str,
        topK: int | None = 20,
        withWeight: bool = False,
        allowPOS: Sequence[str] | tuple[str, ...] = (),
        withFlag: bool = False,
    ) -> list | list[tuple[str, float]]:
        """
        Extract keywords from sentence using TF-IDF algorithm.
        Parameter:
            - topK: return how many top keywords. `None` for all possible words.
            - withWeight: if True, return a list of (word, weight);
                          if False, return a list of words.
            - allowPOS: the allowed POS list eg. ['ns', 'n', 'vn', 'v','nr'].
                        if the POS of w is not in this list,it will be filtered.
            - withFlag: only work with allowPOS is not empty.
                        if True, return a list of pair(word, weight) like posseg.cut
                        if False, return a list of words
        """
        if allowPOS:
            allow_pos_frozen = frozenset(allowPOS)
            words_iter = self.postokenizer.cut(sentence)
        else:
            allow_pos_frozen = None
            words_iter = self.tokenizer.cut(sentence)
        freq: dict[object, float] = {}
        for w in words_iter:  # w may be Pair or str
            if allow_pos_frozen:
                if w.flag not in allow_pos_frozen:
                    continue
                if not withFlag:
                    w = w.word
            wc = w.word if allow_pos_frozen and withFlag else w
            if isinstance(wc, str):
                if len(wc.strip()) < 2 or wc.lower() in self.stop_words:
                    continue
            freq[w] = freq.get(w, 0.0) + 1.0
        total = sum(freq.values())
        for k in list(freq.keys()):
            if allowPOS and withFlag:
                kw = k.word
            elif allowPOS and not withFlag:  # Pair already converted to str
                kw = k
            else:
                kw = k if isinstance(k, str) else k  # may be str
            key_str = kw if isinstance(kw, str) else getattr(kw, "word", str(kw))
            freq[k] *= self.idf_freq.get(key_str, self.median_idf) / (total or 1.0)

        if withWeight:
            tags: list = sorted(freq.items(), key=itemgetter(1), reverse=True)
        else:
            tags = sorted(freq, key=freq.__getitem__, reverse=True)
        if topK:
            return tags[:topK]
        else:
            return tags



import re

from ..jieba_next_rust import _viterbi as viterbi_hmm
from .prob_emit import P as emit_P
from .prob_start import P as start_P
from .prob_trans import P as trans_P

PrevStatus = {"B": "ES", "M": "MB", "S": "SE", "E": "BM"}

Force_Split_Words = set()


def __cut(sentence):
    global emit_P
    prob, pos_list = viterbi_hmm(sentence, "BMES", start_P, trans_P, emit_P)
    begin, nexti = 0, 0
    for i, char in enumerate(sentence):
        pos = pos_list[i]
        if pos == "B":
            begin = i
        elif pos == "E":
            yield sentence[begin : i + 1]
            nexti = i + 1
        elif pos == "S":
            yield char
            nexti = i + 1
    if nexti < len(sentence):
        yield sentence[nexti:]


re_han = re.compile("([\u4e00-\u9fd5]+)")
re_skip = re.compile(r"([a-zA-Z0-9]+(?:\.\d+)?%?)")


def add_force_split(word):
    global Force_Split_Words
    Force_Split_Words.add(word)


def cut(sentence):
    blocks = re_han.split(sentence)
    for blk in blocks:
        if re_han.match(blk):
            for word in __cut(blk):
                if word not in Force_Split_Words:
                    yield word
                else:
                    yield from word
        else:
            tmp = re_skip.split(blk)
            for x in tmp:
                if x:
                    yield x



from __future__ import annotations

from collections import defaultdict
from operator import itemgetter

import jieba_next.posseg

from .tfidf import KeywordExtractor


class UndirectWeightedGraph:
    d: float = 0.85

    def __init__(self):
        self.graph: dict[object, list[tuple[object, object, int]]] = defaultdict(list)

    def addEdge(self, start: object, end: object, weight: int) -> None:
        # use a tuple (start, end, weight) instead of a Edge object
        self.graph[start].append((start, end, weight))
        self.graph[end].append((end, start, weight))

    def rank(self) -> dict[object, float]:
        ws: dict[object, float] = defaultdict(float)
        outSum: dict[object, float] = defaultdict(float)

        wsdef = 1.0 / (len(self.graph) or 1.0)
        for n, out in self.graph.items():
            ws[n] = wsdef
            outSum[n] = sum((e[2] for e in out), 0.0)

        # this line for build stable iteration
        sorted_keys = sorted(self.graph.keys())
        for _ in range(10):  # 10 iters
            for n in sorted_keys:
                s = 0.0
                for e in self.graph[n]:
                    s += e[2] / outSum[e[1]] * ws[e[1]]
                ws[n] = (1 - self.d) + self.d * s

        min_rank, max_rank = float("inf"), float("-inf")

        for w in ws.values():
            if w < min_rank:
                min_rank = w
            if w > max_rank:
                max_rank = w

        for n, w in ws.items():
            # to unify the weights, don't *100.
            ws[n] = (w - min_rank / 10.0) / (max_rank - min_rank / 10.0)

        return ws


class TextRank(KeywordExtractor):
    tokenizer: jieba_next.posseg.POSTokenizer
    postokenizer: jieba_next.posseg.POSTokenizer
    stop_words: set[str]
    pos_filt: frozenset[str]
    span: int

    def __init__(self):
        self.tokenizer = self.postokenizer = jieba_next.posseg.dt
        self.stop_words = self.STOP_WORDS.copy()
        self.pos_filt = frozenset(("ns", "n", "vn", "v"))
        self.span = 5

    def pairfilter(self, wp) -> bool:  # Pair
        return (
            wp.flag in self.pos_filt
            and len(wp.word.strip()) >= 2
            and wp.word.lower() not in self.stop_words
        )

    def textrank(
        self,
        sentence: str,
        topK: int | None = 20,
        withWeight: bool = False,
        allowPOS: tuple[str, ...] = ("ns", "n", "vn", "v"),
        withFlag: bool = False,
    ) -> list:
        """
        Extract keywords from sentence using TextRank algorithm.
        Parameter:
            - topK: return how many top keywords. `None` for all possible words.
            - withWeight: if True, return a list of (word, weight);
                          if False, return a list of words.
            - allowPOS: the allowed POS list eg. ['ns', 'n', 'vn', 'v'].
                        if the POS of w is not in this list, it will be filtered.
            - withFlag: if True, return a list of pair(word, weight) like posseg.cut
                        if False, return a list of words
        """
        self.pos_filt = frozenset(allowPOS)
        g = UndirectWeightedGraph()
        cm: dict[tuple[object, object], int] = defaultdict(int)
        words = tuple(self.tokenizer.cut(sentence))
        for i, wp in enumerate(words):
            if self.pairfilter(wp):
                for j in range(i + 1, i + self.span):
                    if j >= len(words):
                        break
                    if not self.pairfilter(words[j]):
                        continue
                    if allowPOS and withFlag:
                        cm[(wp, words[j])] += 1
                    else:
                        cm[(wp.word, words[j].word)] += 1

        for terms, w in cm.items():
            g.addEdge(terms[0], terms[1], w)
        nodes_rank = g.rank()
        if withWeight:
            tags: list = sorted(nodes_rank.items(), key=itemgetter(1), reverse=True)
        else:
            tags = sorted(nodes_rank, key=nodes_rank.__getitem__, reverse=True)

        if topK:
            return tags[:topK]
        else:
            return tags

    extract_tags = textrank



from __future__ import annotations

import re

import jieba_next

try:
    from whoosh.analysis import (
        LowercaseFilter,
        RegexAnalyzer,
        StemFilter,
        StopFilter,
        Token,
        Tokenizer,
    )
    from whoosh.lang.porter import stem
except Exception:  # whoosh missing
    Token = object
    stem = lambda x: x  # noqa: E731

    class Tokenizer:
        pass


STOP_WORDS = frozenset((
    "a",
    "an",
    "and",
    "are",
    "as",
    "at",
    "be",
    "by",
    "can",
    "for",
    "from",
    "have",
    "if",
    "in",
    "is",
    "it",
    "may",
    "not",
    "of",
    "on",
    "or",
    "tbd",
    "that",
    "the",
    "this",
    "to",
    "us",
    "we",
    "when",
    "will",
    "with",
    "yet",
    "you",
    "your",
    "的",
    "了",
    "和",
))

accepted_chars = re.compile(r"[\u4E00-\u9FD5]+")


class ChineseTokenizer(Tokenizer):
    def __call__(self, text, **kargs):
        words = jieba_next.tokenize(text, mode="search")
        token = Token()
        for w, start_pos, stop_pos in words:
            if not accepted_chars.match(w) and len(w) <= 1:
                continue
            token.original = token.text = w
            token.pos = start_pos
            token.startchar = start_pos
            token.endchar = stop_pos
            yield token


def ChineseAnalyzer(stoplist=STOP_WORDS, minsize=1, stemfn=stem, cachesize=50000):
    return (
        ChineseTokenizer()
        | LowercaseFilter()
        | StopFilter(stoplist=stoplist, minsize=minsize)
        | StemFilter(stemfn=stemfn, ignore=None, cachesize=cachesize)
    )



from setuptools import setup
from setuptools_rust import Binding, RustExtension

setup(
    rust_extensions=[
        RustExtension(
            target="jieba_next.jieba_next_rust",
            path="Cargo.toml",
            binding=Binding.PyO3,
            debug=False,
        )
    ],
    zip_safe=False,
)
